@@ -54,6 +54,8 @@
  * SPDX-License-Identifier: BSD-2-Clause-Patent
  */
 
+#include "config.h"
+
 #include <assert.h>
 #include <errno.h>
 #include <poll.h>
@@ -95,6 +97,7 @@ typedef struct
   bool is_shared;
   int tid;
   int pid;
+  int next_counter_id;
 } SysprofCollector;
 
 #define COLLECTOR_INVALID ((void *)&invalid)
@@ -275,7 +278,7 @@ receive_fd_blocking (int peer_fd)
   if ((cmsg->cmsg_len - ((char *)CMSG_DATA (cmsg) - (char *)cmsg)) % 4 != 0)
     return -1;
 
-  fds = (const int *) CMSG_DATA (cmsg);
+  fds = (const int *)(void *)CMSG_DATA (cmsg);
   n_fds = (cmsg->cmsg_len - ((char *)CMSG_DATA (cmsg) - (char *)cmsg)) / sizeof (*fds);
 
   /* only expecting one FD */
@@ -420,6 +423,7 @@ sysprof_collector_get (void)
 #else
     self->tid = self->pid;
 #endif
+    self->next_counter_id = 1;
 
     pthread_mutex_lock (&control_fd_lock);
 
@@ -572,6 +576,43 @@ sysprof_collector_sample (SysprofBacktraceFunc  backtrace_func,
   } COLLECTOR_END;
 }
 
+void
+sysprof_collector_trace (SysprofBacktraceFunc  backtrace_func,
+                         void                 *backtrace_data,
+                         bool                  entering)
+{
+  COLLECTOR_BEGIN {
+    SysprofCaptureTrace *ev;
+    size_t len;
+
+    len = sizeof *ev + (sizeof (SysprofCaptureTrace) * MAX_UNWIND_DEPTH);
+
+    if ((ev = mapped_ring_buffer_allocate (collector->buffer, len)))
+      {
+        int n_addrs;
+
+        /* See comment from sysprof_collector_allocate(). */
+        if (backtrace_func)
+          n_addrs = backtrace_func (ev->addrs, MAX_UNWIND_DEPTH, backtrace_data);
+        else
+          n_addrs = 0;
+
+        ev->n_addrs = ((n_addrs < 0) ? 0 : (n_addrs > MAX_UNWIND_DEPTH) ? MAX_UNWIND_DEPTH : n_addrs);
+        ev->frame.len = sizeof *ev + sizeof (SysprofCaptureAddress) * ev->n_addrs;
+        ev->frame.type = SYSPROF_CAPTURE_FRAME_TRACE;
+        ev->frame.cpu = _do_getcpu ();
+        ev->frame.pid = collector->pid;
+        ev->frame.time = SYSPROF_CAPTURE_CURRENT_TIME;
+        ev->tid = collector->tid;
+        ev->entering = !!entering;
+        ev->padding1 = 0;
+
+        mapped_ring_buffer_advance (collector->buffer, ev->frame.len);
+      }
+
+  } COLLECTOR_END;
+}
+
 void
 sysprof_collector_mark (int64_t     time,
                         int64_t     duration,
@@ -762,3 +803,114 @@ sysprof_collector_log_printf (int             severity,
       }
   } COLLECTOR_END;
 }
+
+void
+sysprof_collector_define_counters (const SysprofCaptureCounter *counters,
+                                   unsigned int                 n_counters)
+{
+  if (counters == NULL || n_counters == 0)
+    return;
+
+  COLLECTOR_BEGIN {
+    SysprofCaptureCounterDefine *def;
+    size_t len;
+
+    len = realign (sizeof *def + (sizeof *counters * n_counters));
+
+    if ((def = mapped_ring_buffer_allocate (collector->buffer, len)))
+      {
+        def->frame.len = len;
+        def->frame.type = SYSPROF_CAPTURE_FRAME_CTRDEF;
+        def->frame.cpu = _do_getcpu ();
+        def->frame.pid = collector->pid;
+        def->frame.time = SYSPROF_CAPTURE_CURRENT_TIME;
+        def->padding1 = 0;
+        def->padding2 = 0;
+        def->n_counters = n_counters;
+        memcpy (def->counters, counters, sizeof *counters * n_counters);
+
+        mapped_ring_buffer_advance (collector->buffer, def->frame.len);
+      }
+  } COLLECTOR_END;
+}
+
+void
+sysprof_collector_set_counters (const unsigned int               *counters_ids,
+                                const SysprofCaptureCounterValue *values,
+                                unsigned int                      n_counters)
+{
+  if (n_counters == 0)
+    return;
+
+  COLLECTOR_BEGIN {
+    SysprofCaptureCounterSet *set;
+    size_t len;
+    unsigned int n_groups;
+    unsigned int group;
+    unsigned int field;
+    unsigned int i;
+
+    /* Determine how many value groups we need */
+    n_groups = n_counters / SYSPROF_N_ELEMENTS (set->values[0].values);
+    if ((n_groups * SYSPROF_N_ELEMENTS (set->values[0].values)) != n_counters)
+      n_groups++;
+
+    len = realign (sizeof *set + (n_groups * sizeof (SysprofCaptureCounterValues)));
+
+    if ((set = mapped_ring_buffer_allocate (collector->buffer, len)))
+      {
+        set->frame.len = len;
+        set->frame.type = SYSPROF_CAPTURE_FRAME_CTRSET;
+        set->frame.cpu = _do_getcpu ();
+        set->frame.pid = collector->pid;
+        set->frame.time = SYSPROF_CAPTURE_CURRENT_TIME;
+        set->padding1 = 0;
+        set->padding2 = 0;
+        set->n_values = n_groups;
+
+        for (i = 0, group = 0, field = 0; i < n_counters; i++)
+          {
+            set->values[group].ids[field] = counters_ids[i];
+            set->values[group].values[field] = values[i];
+
+            field++;
+
+            if (field == SYSPROF_N_ELEMENTS (set->values[0].values))
+              {
+                field = 0;
+                group++;
+              }
+          }
+
+        mapped_ring_buffer_advance (collector->buffer, set->frame.len);
+      }
+  } COLLECTOR_END;
+}
+
+unsigned int
+sysprof_collector_request_counters (unsigned int n_counters)
+{
+  unsigned int ret = 0;
+
+  if (n_counters == 0)
+    return 0;
+
+  COLLECTOR_BEGIN {
+    ret = collector->next_counter_id;
+    ((SysprofCollector *)collector)->next_counter_id += n_counters;
+  } COLLECTOR_END;
+
+  return ret;
+}
+
+bool
+sysprof_collector_is_active (void)
+{
+  bool ret = false;
+
+  COLLECTOR_BEGIN {
+    ret = true;
+  } COLLECTOR_END;
+
+  return ret;
+}
